/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.util;

import de.bsvrz.pua.prot.util.ExpressionResult.ResultType;

/**
 * Klasse zum Durchführen der arithmetischen Operationen auf long oder double Werten. Wird von {@link de.bsvrz.pua.prot.util.ExpressionTree} verwendet und
 * ermöglicht arithmetische Operationen auf {@link de.bsvrz.pua.prot.util.ExpressionResult} Objekten.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public final class ArithmeticOperation {


	/** Double Wert des linken Operanden */
	private ExpressionResult _left = new ExpressionResult();

	/** Double Wert des rechten Operanden */
	private ExpressionResult _right = new ExpressionResult();
	
	private ExpressionTree _leftTree = null;
	private ExpressionTree _rightTree = null;

	/**
	 * Setzt den linken Operanden
	 *
	 * @param r linker Operand
	 */
	public ArithmeticOperation(final ExpressionResult r) {
		setLeft(r);
	}

	/**
	 * Setzt den linken Operand
	 *
	 * @param longN linker Operand
	 *              
	 * @deprecated Funktion macht keine Unterscheidung bei skalieren Werten              
	 */
	@Deprecated
	public ArithmeticOperation(final long longN) {
		this._left = new ExpressionResult(longN);
	}

	/**
	 * Setzt den linken Operand
	 *
	 * @param doubleN linker Operand
	 */
	public ArithmeticOperation(final double doubleN) {
		this._left = new ExpressionResult(doubleN);
	}

	/**
	 * @param left  Linker Operand.
	 * @param right Rechter Operand.
	 */
	public ArithmeticOperation(final ExpressionResult left, final ExpressionResult right) {
		this(left);
		set(right);
	}

	/**
	 * Setzt den rechten Operand
	 *
	 * @param longN Rechter Operand
	 *              
	 * @deprecated Funktion macht keine Unterscheidugn bei skalierten Werten              
	 */
	@Deprecated
	public void set(final long longN) {
		this._right = new ExpressionResult(longN);
	}

	/**
	 * Setzt den rechten Operand
	 *
	 * @param doubleN Rechter Operand
	 */
	public void set(final double doubleN) {
		this._right = new ExpressionResult(doubleN);
	}

	/**
	 * Setzt den rechten Operand
	 *
	 * @param r Rechter Operand
	 */
	public void set(final ExpressionResult r) {
		this._right = r;
	}

	/**
	 * Prüft &lt;
	 *
	 * @return Ergebnis
	 */
	public boolean less() {
		if(!canCompute()) return fail(ExpressionTree.Operation.less);
		return _left.getDouble() < _right.getDouble();
	}

	private ExpressionResult error(final ExpressionTree.Operation operation) {
		return ExpressionResult.error(ErrorMessageBuilder.illegalOperation(operation, _leftTree, _rightTree, _left, _right));
	}

	private boolean fail(final ExpressionTree.Operation operation) {
		throw new IllegalArgumentException(ErrorMessageBuilder.illegalOperation(operation, _leftTree, _rightTree, _left, _right));
	}

	private boolean canCompute() {
		return _left.isArithmetic() && _right.isArithmetic();
	}

	/**
	 * Prüft &lt;=
	 *
	 * @return Ergebnis
	 */
	public boolean lessEqual() {
		if(!canCompute()) return fail(ExpressionTree.Operation.lessEqual);
		return _left.getDouble() <= _right.getDouble();
	}

	/**
	 * Prüft &gt;
	 *
	 * @return Ergebnis
	 */
	public boolean greater() {
		if(!canCompute()) return fail(ExpressionTree.Operation.greater);
		return _left.getDouble() > _right.getDouble();
	}

	/**
	 * Prüft &gt;=
	 *
	 * @return Ergebnis
	 */
	public boolean greaterEqual() {
		if(!canCompute()) return fail(ExpressionTree.Operation.greaterEqual);
		return _left.getDouble() >= _right.getDouble();
	}

	/**
	 * Prüft auf Gleichheit der Operanden.
	 *
	 * @return Ergebnis
	 */
	public boolean equal() {
		if(!canCompute()) return fail(ExpressionTree.Operation.equal);
		return _left.getDouble() == _right.getDouble();
	}

	/**
	 * Prüft auf Ungleichheit der Operanden.
	 *
	 * @return Ergebnis
	 */
	public boolean notEqual() {
		if(!canCompute()) return fail(ExpressionTree.Operation.notEqual);
		return _left.getDouble() != _right.getDouble();
	}

	/**
	 * Subtrahiert die Operanden. Falls einer der Operanden vom Typ double ist, ist das Ergebnis auch vom Typ double.
	 *
	 * @return Ergebnis der Operation
	 */
	public ExpressionResult performMinusOperation() {
		if(!canCompute()) return error(ExpressionTree.Operation.subtract);
		return new ExpressionResult(_left.getDouble() - _right.getDouble());
	}

	/**
	 * Addiert die Operanden. Falls einer der Operanden vom Typ double ist, ist das Ergebnis auch vom Typ double.
	 *
	 * @return Ergebnis der Operation
	 */
	public ExpressionResult performPlusOperation() {
		if(!canCompute()) return error(ExpressionTree.Operation.add);
		return new ExpressionResult(_left.getDouble() + _right.getDouble());
	}

	/**
	 * Multipliziert die Operanden. Falls einer der Operanden vom Typ double ist, ist das Ergebnis auch vom Typ double.
	 *
	 * @return Ergebnis der Operation
	 */
	public ExpressionResult performMultOperation() {
		if(!canCompute()) return error(ExpressionTree.Operation.mult);
		return new ExpressionResult(_left.getDouble() * _right.getDouble());
	}

	/**
	 * Dividiert die Operanden. Falls einer der Operanden vom Typ double ist, ist das Ergebnis auch vom Typ double.
	 *
	 * @return Ergebnis der Operation
	 */
	public ExpressionResult performDivOperation() {
		if(!canCompute()) return error(ExpressionTree.Operation.div);
		double a = _left.getDouble();
		double b = _right.getDouble();
		if(b == 0){
			return ExpressionResult.error(ErrorMessageBuilder.divisionByZero(ExpressionTree.Operation.div, _leftTree, _rightTree, _left, _right));
		}
		return new ExpressionResult(a / b);
	}

	/**
	 * Ganzzahl-Division der Operanden. Das Ergebnis ist immer vom Typ long.
	 *
	 * @return Ergebnis der Operation
	 */
	public ExpressionResult performDivIntOperation() {
		if(!canCompute()) return error(ExpressionTree.Operation.divInt);
		long a = (long) _left.getDouble();
		long b = (long) _right.getDouble();
		if(b == 0){
			return ExpressionResult.error(ErrorMessageBuilder.divisionByZero(ExpressionTree.Operation.divInt, _leftTree, _rightTree, _left, _right));
		}
		return new ExpressionResult(a / b);
	}

	/**
	 * Modulo-Operation
	 *
	 * @return Ergebnis der Operation
	 */
	public ExpressionResult performModOperation() {
		if(!canCompute()) return error(ExpressionTree.Operation.modulo);
		double a = _left.getDouble();
		double b = _right.getDouble();
		if(b == 0) {
			return ExpressionResult.error(ErrorMessageBuilder.divisionByZero(ExpressionTree.Operation.modulo, _leftTree, _rightTree, _left, _right));
		}
		return new ExpressionResult(a % b);
	}

	/**
	 * Setzt den linken Operanden
	 *
	 * @param leftOperand Linker Operand
	 */
	public void setLeft(final ExpressionResult leftOperand) {
		_left = leftOperand;
	}

	public void setLeftTree(final ExpressionTree leftTree) {
		_leftTree = leftTree;
	}

	public void setRightTree(final ExpressionTree rightTree) {
		_rightTree = rightTree;
	}

	/** @see Object#toString() */
	@Override
	public String toString() {
		final StringBuilder sb = new StringBuilder();

		sb.append(_left);

		sb.append(":");

		sb.append(_right);

		return sb.toString();
	}
}
